# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 Compatible Cloud Storage, (C) 2016 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.copy_conditions
~~~~~~~~~~~~~~~

This module contains :class:`CopyConditions <CopyConditions>` implementation.

:copyright: (c) 2016 by Minio, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

import collections
from .helpers import is_non_empty_string

# CopyCondition explanation:
# http://docs.aws.amazon.com/AmazonS3/latest/API/RESTObjectCOPY.html
#
# Example:
#
#  copyCondition {
#      key: "x-amz-copy-if-modified-since",
#      value: "Tue, 15 Nov 1994 12:45:26 GMT",
#
class CopyConditions(collections.MutableMapping):
    """
    A :class:`CopyConditions <CopyConditions>` collection of
       supported CopyObject conditions.

        - x-amz-copy-source-if-match
        - x-amz-copy-source-if-none-match
        - x-amz-copy-source-if-unmodified-since
        - x-amz-copy-source-if-modified-since

    """
    def __init__(self, *args, **kwargs):
        self._store = dict()
        self.update(dict(*args, **kwargs))  # use the free update to set keys

    def __getitem__(self, key):
        return self._store[key]

    def __setitem__(self, key, value):
        self._store[key] = value

    def __delitem__(self, key):
        del self._store[key]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def set_match_etag(self, etag):
        """
        """
        is_non_empty_string(etag)
        self._store['X-Amz-Copy-Source-If-Match'] = etag

    def set_match_etag_except(self, etag):
        """
        """
        is_non_empty_string(etag)
        self._store['X-Amz-Copy-Source-If-None-Match'] = etag

    def set_unmodified_since(self, mod_time):
        """
        """
        time = mod_time.strftime('%a, %d %b %Y %H:%M:%S GMT')
        self._store['X-Amz-Copy-Source-If-Unmodified-Since'] = time

    def set_modified_since(self, mod_time):
        """
        """
        time = mod_time.strftime('%a, %d %b %Y %H:%M:%S GMT')
        self._store['X-Amz-Copy-Source-If-Modified-Since'] = time



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2015, 2016, 2017 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.helpers

This module implements all helper functions.

:copyright: (c) 2015, 2016, 2017 by Minio, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from __future__ import absolute_import
# if math.ceil returns an integer and devide two integers returns a float, calculate
# part size will cause errors, so make sure division integers returns a float.
from __future__ import division
import io

import collections
import base64
import hashlib
import re
import os
import errno
import math

from .compat import (urlsplit, urlencode, queryencode,
                     str, bytes, basestring, _is_py3, _is_py2)
from .error import (InvalidBucketError, InvalidEndpointError,
                    InvalidArgumentError)

# Constants
MAX_MULTIPART_COUNT = 10000 # 10000 parts
MAX_MULTIPART_OBJECT_SIZE = 5 * 1024 * 1024 * 1024 * 1024  # 5TiB
MAX_POOL_SIZE = 10
MIN_PART_SIZE = 5 * 1024 * 1024  # 5MiB

_VALID_BUCKETNAME_REGEX = re.compile('^[a-z0-9][a-z0-9\\.\\-]+[a-z0-9]$')
_ALLOWED_HOSTNAME_REGEX = re.compile(
    '^((?!-)[A-Z\\d-]{1,63}(?<!-)\\.)*((?!-)[A-Z\\d-]{1,63}(?<!-))$',
    re.IGNORECASE)

_EXTRACT_REGION_REGEX = re.compile('s3[.-]?(.+?).amazonaws.com')

def get_s3_region_from_endpoint(endpoint):
    """
    Extracts and returns an AWS S3 region from an endpoint
    of form `s3-ap-southeast-1.amazonaws.com`

    :param endpoint: Endpoint region to be extracted.
    """

    # Extract region by regex search.
    m = _EXTRACT_REGION_REGEX.search(endpoint)
    if m:
        # Regex matches, we have found a region.
        region = m.group(1)
        if region == 'external-1':
            # Handle special scenario for us-east-1 URL.
            return 'us-east-1'
        if region.startswith('dualstack'):
            # Handle special scenario for dualstack URL.
            return region.split('.')[1]
        return region

    # No regex matches return None.
    return None

def dump_http(method, url, request_headers, response, output_stream):
    """
    Dump all headers and response headers into output_stream.

    :param request_headers: Dictionary of HTTP request headers.
    :param response_headers: Dictionary of HTTP response headers.
    :param output_stream: Stream where the request is being dumped at.
    """

    # Start header.
    output_stream.write('---------START-HTTP---------\n')

    # Get parsed url.
    parsed_url = urlsplit(url)

    # Dump all request headers recursively.
    http_path = parsed_url.path
    if parsed_url.query:
        http_path = http_path + '?' + parsed_url.query

    output_stream.write('{0} {1} HTTP/1.1\n'.format(method,
                                                    http_path))

    for k, v in list(request_headers.items()):
        if k is 'authorization':
            # Redact signature header value from trace logs.
            v = re.sub(r'Signature=([[0-9a-f]+)', 'Signature=*REDACTED*', v)
        output_stream.write('{0}: {1}\n'.format(k.title(), v))

    # Write a new line.
    output_stream.write('\n')

    # Write response status code.
    output_stream.write('HTTP/1.1 {0}\n'.format(response.status))

    # Dump all response headers recursively.
    for k, v in list(response.getheaders().items()):
        output_stream.write('{0}: {1}\n'.format(k.title(), v))

    # For all errors write all the available response body.
    if response.status != 200 and \
       response.status != 204 and response.status != 206:
        output_stream.write('{0}'.format(response.read()))

    # End header.
    output_stream.write('---------END-HTTP---------\n')

def mkdir_p(path):
    """
    Recursively creates parent and sub directories.

    :param path:
    """
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

class PartMetadata(object):
    """
    Parts manager split parts metadata :class:`PartMetadata <PartMetadata>`.

    :param data: Part writer object backed by temporary file.
    :param md5_hex: MD5 hash in hex format.
    :param sha256_hex: Sha256 hash in hex format.
    :param size: Size of the part.
    """
    def __init__(self, data, md5_hex, sha256_hex, size):
        self.data = data
        self.md5_hex = md5_hex
        self.sha256_hex = sha256_hex
        self.size = size

def read_full(data, size):
    """
    read_full reads exactly `size` bytes from reader. returns
    `size` bytes.

    :param data: Input stream to read from.
    :param size: Number of bytes to read from `data`.
    :return: Returns :bytes:`part_data`
    """
    default_read_size = 32768 # 32KiB per read operation.
    chunk = io.BytesIO()
    chunk_size = 0

    while chunk_size < size:
        read_size = default_read_size
        if (size - chunk_size) < default_read_size:
            read_size = size - chunk_size
        current_data = data.read(read_size)
        if not current_data or len(current_data) == 0:
            break
        chunk.write(current_data)
        chunk_size+= len(current_data)

    return chunk.getvalue()

AWS_S3_ENDPOINT_MAP = {
    'us-east-1': 's3.amazonaws.com',
    'us-east-2': 's3-us-east-2.amazonaws.com',
    'us-west-2': 's3-us-west-2.amazonaws.com',
    'us-west-1': 's3-us-west-1.amazonaws.com',
    'ca-central-1': 's3.ca-central-1.amazonaws.com',
    'eu-west-1': 's3-eu-west-1.amazonaws.com',
    'eu-west-2': 's3-eu-west-2.amazonaws.com',
    'sa-east-1': 's3-sa-east-1.amazonaws.com',
    'eu-central-1': 's3-eu-central-1.amazonaws.com',
    'ap-south-1': 's3.ap-south-1.amazonaws.com',
    'ap-southeast-1': 's3-ap-southeast-1.amazonaws.com',
    'ap-southeast-2': 's3-ap-southeast-2.amazonaws.com',
    'ap-northeast-1': 's3-ap-northeast-1.amazonaws.com',
    'ap-northeast-2': 's3-ap-northeast-2.amazonaws.com',
    'cn-north-1': 's3.cn-north-1.amazonaws.com.cn'
}

def get_s3_endpoint(region):
    return AWS_S3_ENDPOINT_MAP.get(region, 's3.amazonaws.com')

def get_target_url(endpoint_url, bucket_name=None, object_name=None,
                   bucket_region='us-east-1', query=None):
    """
    Construct final target url.

    :param endpoint_url: Target endpoint url where request is served to.
    :param bucket_name: Bucket component for the target url.
    :param object_name: Object component for the target url.
    :param bucket_region: Bucket region for the target url.
    :param query: Query parameters as a *dict* for the target url.
    :return: Returns final target url as *str*.
    """
    # New url
    url = None

    # Parse url
    parsed_url = urlsplit(endpoint_url)

    # Get new host, scheme.
    scheme = parsed_url.scheme
    host = parsed_url.netloc

    # Strip 80/443 ports since curl & browsers do not
    # send them in Host header.
    if (scheme == 'http' and parsed_url.port == 80) or\
       (scheme == 'https' and parsed_url.port == 443):
        host = parsed_url.hostname

    if 's3.amazonaws.com' in host:
        host = get_s3_endpoint(bucket_region)

    url = scheme + '://' + host
    if bucket_name:
        # Save if target url will have buckets which suppport
        # virtual host.
        is_virtual_host_style = is_virtual_host(endpoint_url,
                                                bucket_name)
        if is_virtual_host_style:
            url = (scheme + '://' + bucket_name + '.' + host)
        else:
            url = (scheme + '://' + host + '/' + bucket_name)

    url_components = [url]
    url_components.append('/')

    if object_name:
        object_name = encode_object_name(object_name)
        url_components.append(object_name)

    if query:
        ordered_query = collections.OrderedDict(sorted(query.items()))
        query_components = []
        for component_key in ordered_query:
            if ordered_query[component_key] is not None:
                if isinstance(ordered_query[component_key], list):
                    for value in ordered_query[component_key]:
                        query_components.append(component_key+'='+
                                                queryencode(value))
                else:
                    query_components.append(
                        component_key+'='+
                        queryencode(
                            ordered_query[component_key]
                        )
                    )
            else:
                query_components.append(component_key)

        query_string = '&'.join(query_components)
        if query_string:
            url_components.append('?')
            url_components.append(query_string)

    return ''.join(url_components)


def is_valid_endpoint(endpoint):
    """
    Verify if endpoint is valid.

    :type endpoint: string
    :param endpoint: An endpoint. Must have at least a scheme and a hostname.
    :return: True if the endpoint is valid. Raise :exc:`InvalidEndpointError`
       otherwise.
    """
    try:
        if urlsplit(endpoint).scheme:
            raise InvalidEndpointError('Hostname cannot have a scheme.')

        hostname = endpoint.split(':')[0]
        if hostname is None:
            raise InvalidEndpointError('Hostname cannot be empty.')

        if len(hostname) > 255:
            raise InvalidEndpointError('Hostname cannot be greater than 255.')

        if hostname[-1] == '.':
            hostname = hostname[:-1]

        if not _ALLOWED_HOSTNAME_REGEX.match(hostname):
            raise InvalidEndpointError('Hostname does not meet URL standards.')

    except AttributeError as error:
        raise TypeError(error)

    return True

def is_virtual_host(endpoint_url, bucket_name):
    """
    Check to see if the ``bucket_name`` can be part of virtual host
    style.

    :param endpoint_url: Endpoint url which will be used for virtual host.
    :param bucket_name: Bucket name to be validated against.
    """
    is_valid_bucket_name(bucket_name)

    parsed_url = urlsplit(endpoint_url)
    # bucket_name can be valid but '.' in the hostname will fail
    # SSL certificate validation. So do not use host-style for
    # such buckets.
    if 'https' in parsed_url.scheme and '.' in bucket_name:
        return False
    for host in ['s3.amazonaws.com', 'aliyuncs.com']:
        if host in parsed_url.netloc:
            return True
    return False

def is_valid_bucket_name(bucket_name):
    """
    Check to see if the ``bucket_name`` complies with the
    restricted DNS naming conventions necessary to allow
    access via virtual-hosting style.

    :param bucket_name: Bucket name in *str*.
    :return: True if the bucket is valid. Raise :exc:`InvalidBucketError`
       otherwise.
    """
    # Verify bucket name length.
    if len(bucket_name) < 3:
        raise InvalidBucketError('Bucket name cannot be less than'
                                 ' 3 characters.')
    if len(bucket_name) > 63:
        raise InvalidBucketError('Bucket name cannot be more than'
                                 ' 63 characters.')
    if '..' in bucket_name:
        raise InvalidBucketError('Bucket name cannot have successive'
                                 ' periods.')

    match = _VALID_BUCKETNAME_REGEX.match(bucket_name)
    if match is None or match.end() != len(bucket_name):
        raise InvalidBucketError('Bucket name does not follow S3 standards.'
                                 ' Bucket: {0}'.format(bucket_name))

    return True


def is_non_empty_string(input_string):
    """
    Validate if non empty string

    :param input_string: Input is a *str*.
    :return: True if input is string and non empty.
       Raise :exc:`Exception` otherwise.
    """
    try:
        if not input_string.strip():
            raise ValueError()
    except AttributeError as error:
        raise TypeError(error)

    return True

def is_valid_policy_type(policy):
    """
    Validate if policy is type str

    :param policy: S3 style Bucket policy.
    :return: True if policy parameter is of a valid type, 'string'.
    Raise :exc:`TypeError` otherwise.
    """
    if _is_py3:
        string_type = str,
    elif _is_py2:
        string_type = basestring

    if not isinstance(policy, string_type):
        raise TypeError('policy can only be of type str')

    is_non_empty_string(policy)

    return True

def is_valid_bucket_notification_config(notifications):
    """
    Validate the notifications config structure

    :param notifications: Dictionary with specific structure.
    :return: True if input is a valid bucket notifications structure.
       Raise :exc:`InvalidArgumentError` otherwise.
    """
    # check if notifications is a dict.
    if not isinstance(notifications, dict):
        raise TypeError('notifications configuration must be a dictionary')

    if len(notifications) == 0:
        raise InvalidArgumentError(
            'notifications configuration may not be empty'
        )

    VALID_NOTIFICATION_KEYS = set([
        "TopicConfigurations",
        "QueueConfigurations",
        "CloudFunctionConfigurations",
    ])

    VALID_SERVICE_CONFIG_KEYS = set([
        'Id',
        'Arn',
        'Events',
        'Filter',
    ])

    NOTIFICATION_EVENTS = set([
        's3:ReducedRedundancyLostObject',
        's3:ObjectCreated:*',
        's3:ObjectCreated:Put',
        's3:ObjectCreated:Post',
        's3:ObjectCreated:Copy',
        's3:ObjectCreated:CompleteMultipartUpload',
        's3:ObjectRemoved:*',
        's3:ObjectRemoved:Delete',
        's3:ObjectRemoved:DeleteMarkerCreated',
    ])

    for key, value in notifications.items():
        # check if key names are valid
        if key not in VALID_NOTIFICATION_KEYS:
            raise InvalidArgumentError((
                '{} is an invalid key '
                'for notifications configuration').format(key))

        # check if config values conform
        # first check if value is a list
        if not isinstance(value, list):
            raise InvalidArgumentError((
                'The value for key "{}" in the notifications '
                'configuration must be a list.').format(key))

        for service_config in value:
            # check type matches
            if not isinstance(service_config, dict):
                raise InvalidArgumentError((
                    'Each service configuration item for "{}" must be a '
                    'dictionary').format(key))

            # check keys are valid
            for skey in service_config.keys():
                if skey not in VALID_SERVICE_CONFIG_KEYS:
                    raise InvalidArgumentError((
                        '{} is an invalid key for a service '
                        'configuration item').format(skey))

            # check for required keys
            arn = service_config.get('Arn', '')
            if arn == '':
                raise InvalidArgumentError(
                    'Arn key in service config must be present and has to be '
                    'non-empty string'
                )
            events = service_config.get('Events', [])
            if len(events) < 1:
                raise InvalidArgumentError(
                    'At least one event must be specified in a service config'
                )
            if not isinstance(events, list):
                raise InvalidArgumentError('"Events" must be a list of strings '
                                           'in a service configuration')

            # check if 'Id' key is present, it should be string or bytes.
            if not isinstance(service_config.get('Id', ''), basestring):
                raise InvalidArgumentError('"Id" key must be a string')

            for event in events:
                if event not in NOTIFICATION_EVENTS:
                    raise InvalidArgumentError(
                        '{} is not a valid event. Valid '
                        'events are: {}'.format(event, NOTIFICATION_EVENTS))

            if 'Filter' in service_config:
                exception_msg = (
                    '{} - If a Filter key is given, it must be a '
                    'dictionary, the dictionary must have the '
                    'key "Key", and its value must be an object, with '
                    'a key named "FilterRules" which must be a non-empty list.'
                ).format(
                    service_config['Filter']
                )
                try:
                    filter_rules = service_config.get('Filter', {}).get(
                        'Key', {}).get('FilterRules', [])
                    if not isinstance(filter_rules, list):
                        raise InvalidArgumentError(exception_msg)
                    if len(filter_rules) < 1:
                        raise InvalidArgumentError(exception_msg)
                except AttributeError:
                    raise InvalidArgumentError(exception_msg)
                for filter_rule in filter_rules:
                    try:
                        name = filter_rule['Name']
                        value = filter_rule['Value']
                    except KeyError:
                        raise InvalidArgumentError((
                            '{} - a FilterRule dictionary must have "Name" '
                             'and "Value" keys').format(filter_rule))

                    if name not in ['prefix', 'suffix']:
                        raise InvalidArgumentError((
                            '{} - The "Name" key in a filter rule must be '
                             'either "prefix" or "suffix"').format(name))

    return True

def is_valid_sse_c_object(sse=None):
    """
    Validate the SSE object and type

    :param sse: SSE object defined.
    """
    if sse and sse.type() != "SSE-C":
            raise InvalidArgumentError("Required type SSE-C object to be passed")

def is_valid_sse_object(sse):
    """
    Validate the SSE object and type

    :param sse: SSE object defined.
    """
    if sse and sse.type() != "SSE-C" and sse.type() != "SSE-KMS" and sse.type() != "SSE-S3":
        raise InvalidArgumentError("unsuported type of sse argument in put_object")

def is_valid_source_sse_object(sse):
    """
    Validate the SSE object and type

    :param sse: SSE object defined.
    """
    if sse and sse.type() != "copy_SSE-C":
        raise InvalidArgumentError("Required type copy_SSE-C object to be passed")

def encode_object_name(object_name):
    """
    URL encode input object name.

    :param object_name: Un-encoded object name.
    :return: URL encoded input object name.
    """
    is_non_empty_string(object_name)
    return urlencode(object_name)

class Hasher(object):
    """
    Adaptation of hashlib-based hash functions that
    return unicode-encoded hex- and base64-digest
    strings.
    """
    def __init__(self, data, h):
        if data is None:
            data = b''
        if isinstance(data, str):
            data = data.encode('utf-8')
        self.h = h(data)

    @classmethod
    def md5(cls, data=''):
        return cls(data, hashlib.md5)

    @classmethod
    def sha256(cls, data=''):
        return cls(data, hashlib.sha256)

    def update(self, data):
        if isinstance(data, str):
            data = data.encode('utf-8')
        self.h.update(data)

    def hexdigest(self):
        r = self.h.hexdigest()
        return r.decode('utf-8') if isinstance(r, bytes) else r

    def base64digest(self):
        r = base64.b64encode(self.h.digest())
        return r.decode('utf-8') if isinstance(r, bytes) else r


def get_sha256_hexdigest(content):
    """
    Calculate sha256 hexdigest of content.

    :param content: Input str or bytes. If the type is `str` we encode
    it to UTF8 first.

    :return: sha256 digest encoded as hexadecimal `str`.

    """
    return Hasher.sha256(content).hexdigest()


def get_md5_base64digest(content):
    """Calculate md5sum and return digest as base64 encoded string.

    :param content: Input str or bytes. If the type is `str` we encode
    it to UTF8 and calculate md5sum.

    :return: md5 digest encoded to base64 `str`.

    """
    return Hasher.md5(content).base64digest()


def optimal_part_info(length):
    """
    Calculate optimal part size for multipart uploads.

    :param length: Input length to calculate part size of.
    :return: Optimal part size.
    """
    # object size is '-1' set it to 5TiB.
    if length == -1:
        length = MAX_MULTIPART_OBJECT_SIZE
    if length > MAX_MULTIPART_OBJECT_SIZE:
        raise InvalidArgumentError('Input content size is bigger '
                                   ' than allowed maximum of 5TiB.')

    # Use floats for part size for all calculations to avoid
    # overflows during float64 to int64 conversions.
    part_size_float = math.ceil(length/MAX_MULTIPART_COUNT)
    part_size_float = (math.ceil(part_size_float/MIN_PART_SIZE)
                       * MIN_PART_SIZE)
    # Total parts count.
    total_parts_count = int(math.ceil(length/part_size_float))
    # Part size.
    part_size = int(part_size_float)
    # Last part size.
    last_part_size = length - int(total_parts_count-1)*part_size
    return total_parts_count, part_size, last_part_size

# return a new metadata dictionary where user defined metadata keys
# are prefixed by "x-amz-meta-"
def amzprefix_user_metadata(metadata):
    m = dict()
    for k,v in metadata.items():
       if is_amz_header(k) or is_supported_header(k) or is_storageclass_header(k):
            m[k] = v
       else:
            m["X-Amz-Meta-" + k] = v
    return m

# returns true if amz s3 system defined metadata
def is_amz_header(key):
    key = key.lower()
    return key.startswith("x-amz-meta") or key == "x-amz-acl" or key.startswith("x-amz-server-side-encryption")

# returns true if a standard supported header
def is_supported_header(key):
    ## Supported headers for object.
    supported_headers = [
	   "cache-control",
	   "content-encoding",
	   "content-disposition",
	   "content-language",
	   "x-amz-website-redirect-location",
            ## Add more supported headers here.
        ]
    return key.lower() in supported_headers

# returns true if header is a storage class header
def is_storageclass_header(key):
    return key.lower() == "x-amz-storage-class"



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 Compatible Cloud Storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.parsers
~~~~~~~~~~~~~~~~~~~

This module contains core API parsers.

:copyright: (c) 2015 by Minio, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

# standard.
from xml.etree import cElementTree
from xml.etree.cElementTree import ParseError

from datetime import datetime

# dependencies.
import pytz

# minio specific.
from .error import (InvalidXMLError, MultiDeleteError)
from .compat import urldecode
from .definitions import (Object, Bucket, IncompleteUpload,
                          UploadPart, MultipartUploadResult,
                          CopyObjectResult)
from .xml_marshal import (NOTIFICATIONS_ARN_FIELDNAME_MAP)


if hasattr(cElementTree, 'ParseError'):
    _ETREE_EXCEPTIONS = (ParseError, AttributeError, ValueError, TypeError)
else:
    _ETREE_EXCEPTIONS = (SyntaxError, AttributeError, ValueError, TypeError)

_S3_NS = {'s3' : 'http://s3.amazonaws.com/doc/2006-03-01/'}


class S3Element(object):
    """S3 aware XML parsing class. Wraps a root element name and
    cElementTree.Element instance. Provides S3 namespace aware parsing
    functions.

    """
    def __init__(self, root_name, element):
        self.root_name = root_name
        self.element = element

    @classmethod
    def fromstring(cls, root_name, data):
        """Initialize S3Element from name and XML string data.

        :param name: Name for XML data. Used in XML errors.
        :param data: string data to be parsed.
        :return: Returns an S3Element.
        """
        try:
            return cls(root_name, cElementTree.fromstring(data))
        except _ETREE_EXCEPTIONS as error:
            raise InvalidXMLError(
                '"{}" XML is not parsable. Message: {}'.format(
                    root_name, error.message
                )
            )

    def findall(self, name):
        """Similar to ElementTree.Element.findall()

        """
        return [
            S3Element(self.root_name, elem)
            for elem in self.element.findall('s3:{}'.format(name), _S3_NS)
        ]

    def find(self, name):
        """Similar to ElementTree.Element.find()

        """
        elt = self.element.find('s3:{}'.format(name), _S3_NS)
        return S3Element(self.root_name, elt) if elt is not None else None

    def get_child_text(self, name, strict=True):
        """Extract text of a child element. If strict, and child element is
        not present, raises InvalidXMLError and otherwise returns
        None.

        """
        if strict:
            try:
                return self.element.find('s3:{}'.format(name), _S3_NS).text
            except _ETREE_EXCEPTIONS as error:
                raise InvalidXMLError(
                    ('Invalid XML provided for "{}" - erroring tag <{}>. '
                     'Message: {}').format(self.root_name, name, error.message)
                )
        else:
            return self.element.findtext('s3:{}'.format(name), None, _S3_NS)

    def get_urldecoded_elem_text(self, name, strict=True):
        """Like self.get_child_text(), but also performs urldecode() on the
        result.

        """
        text = self.get_child_text(name, strict)
        # strictness is already enforced above.
        return urldecode(text) if text is not None else None

    def get_etag_elem(self, strict=True):
        """Fetches an 'ETag' child element suitably processed.

        """
        return self.get_child_text('ETag', strict).replace('"', '')

    def get_int_elem(self, name):
        """Fetches an integer type XML child element by name.

        """
        return int(self.get_child_text(name))

    def get_localized_time_elem(self, name):
        """Parse a time XML child element.

        """
        return _iso8601_to_localized_time(self.get_child_text(name))

    def text(self):
        """Fetch the current node's text

        """
        return self.element.text

    def is_dir(self):
        """Returns True if the object is a dir
        ie, if an object name has `/` suffixed.

        """
        text = self.get_child_text('Key')
        return text.endswith("/")


def parse_multipart_upload_result(data):
    """
    Parser for complete multipart upload response.

    :param data: Response data for complete multipart upload.
    :return: :class:`MultipartUploadResult <MultipartUploadResult>`.
    """
    root = S3Element.fromstring('CompleteMultipartUploadResult', data)

    return MultipartUploadResult(
        root.get_child_text('Bucket'),
        root.get_child_text('Key'),
        root.get_child_text('Location'),
        root.get_etag_elem()
    )

def parse_copy_object(bucket_name, object_name, data):
    """
    Parser for copy object response.

    :param data: Response data for copy object.
    :return: :class:`CopyObjectResult <CopyObjectResult>`
    """
    root = S3Element.fromstring('CopyObjectResult', data)

    return CopyObjectResult(
        bucket_name, object_name,
        root.get_etag_elem(),
        root.get_localized_time_elem('LastModified')
    )

def parse_list_buckets(data):
    """
    Parser for list buckets response.

    :param data: Response data for list buckets.
    :return: List of :class:`Bucket <Bucket>`.
    """
    root = S3Element.fromstring('ListBucketsResult', data)

    return [
        Bucket(bucket.get_child_text('Name'),
               bucket.get_localized_time_elem('CreationDate'))
        for buckets in root.findall('Buckets')
        for bucket in buckets.findall('Bucket')
    ]

def _parse_objects_from_xml_elts(bucket_name, contents, common_prefixes):
    """Internal function that extracts objects and common prefixes from
    list_objects responses.
    """
    objects = [
        Object(bucket_name,
               content.get_child_text('Key'),
               content.get_localized_time_elem('LastModified'),
               content.get_etag_elem(strict=False),
               content.get_int_elem('Size'),
               is_dir=content.is_dir())
        for content in contents
    ]

    object_dirs = [
        Object(bucket_name, dir_elt.text(), None, '',
               0, is_dir=True)
        for dirs_elt in common_prefixes
        for dir_elt in dirs_elt.findall('Prefix')
    ]

    return objects, object_dirs

def parse_list_objects(data, bucket_name):
    """
    Parser for list objects response.

    :param data: Response data for list objects.
    :param bucket_name: Response for the bucket.
    :return: Replies back three distinctive components.
       - List of :class:`Object <Object>`
       - True if list is truncated, False otherwise.
       - Object name marker for the next request.
    """
    root = S3Element.fromstring('ListObjectResult', data)

    is_truncated = root.get_child_text('IsTruncated').lower() == 'true'
    # NextMarker element need not be present.
    marker = root.get_urldecoded_elem_text('NextMarker', strict=False)
    objects, object_dirs = _parse_objects_from_xml_elts(
        bucket_name,
        root.findall('Contents'),
        root.findall('CommonPrefixes')
    )

    if is_truncated and marker is None:
        marker = objects[-1].object_name

    return objects + object_dirs, is_truncated, marker


def parse_list_objects_v2(data, bucket_name):
    """
    Parser for list objects version 2 response.

    :param data: Response data for list objects.
    :param bucket_name: Response for the bucket.
    :return: Returns three distinct components:
       - List of :class:`Object <Object>`
       - True if list is truncated, False otherwise.
       - Continuation Token for the next request.
    """
    root = S3Element.fromstring('ListObjectV2Result', data)

    is_truncated = root.get_child_text('IsTruncated').lower() == 'true'
    # NextContinuationToken may not be present.
    continuation_token = root.get_child_text('NextContinuationToken',
                                             strict=False)
    objects, object_dirs = _parse_objects_from_xml_elts(
        bucket_name,
        root.findall('Contents'),
        root.findall('CommonPrefixes')
    )

    return objects + object_dirs, is_truncated, continuation_token

def parse_list_multipart_uploads(data, bucket_name):
    """
    Parser for list multipart uploads response.

    :param data: Response data for list multipart uploads.
    :param bucket_name: Response for the bucket.
    :return: Replies back four distinctive components.
       - List of :class:`IncompleteUpload <IncompleteUpload>`
       - True if list is truncated, False otherwise.
       - Object name marker for the next request.
       - Upload id marker for the next request.
    """
    root = S3Element.fromstring('ListMultipartUploadsResult', data)

    is_truncated = root.get_child_text('IsTruncated').lower() == 'true'
    key_marker = root.get_urldecoded_elem_text('NextKeyMarker', strict=False)
    upload_id_marker = root.get_child_text('NextUploadIdMarker', strict=False)
    uploads = [
        IncompleteUpload(bucket_name,
                         upload.get_urldecoded_elem_text('Key'),
                         upload.get_child_text('UploadId'),
                         upload.get_localized_time_elem('Initiated'))
        for upload in root.findall('Upload')
    ]

    return uploads, is_truncated, key_marker, upload_id_marker


def parse_list_parts(data, bucket_name, object_name, upload_id):
    """
    Parser for list parts response.

    :param data: Response data for list parts.
    :param bucket_name: Response for the bucket.
    :param object_name: Response for the object.
    :param upload_id: Upload id of object name for
       the active multipart session.
    :return: Replies back three distinctive components.
       - List of :class:`UploadPart <UploadPart>`.
       - True if list is truncated, False otherwise.
       - Next part marker for the next request if the
         list was truncated.
    """
    root = S3Element.fromstring('ListPartsResult', data)

    is_truncated = root.get_child_text('IsTruncated').lower() == 'true'
    part_marker = root.get_child_text('NextPartNumberMarker', strict=False)
    parts = [
        UploadPart(bucket_name, object_name, upload_id,
                   part.get_int_elem('PartNumber'),
                   part.get_etag_elem(),
                   part.get_localized_time_elem('LastModified'),
                   part.get_int_elem('Size'))
        for part in root.findall('Part')
    ]

    return parts, is_truncated, part_marker

def parse_new_multipart_upload(data):
    """
    Parser for new multipart upload response.

    :param data: Response data for new multipart upload.
    :return: Returns a upload id.
    """
    root = S3Element.fromstring('InitiateMultipartUploadResult', data)
    return root.get_child_text('UploadId')

def parse_location_constraint(data):
    """
    Parser for location constraint response.

    :param data: Response data for get bucket location.
    :return: Returns location of your bucket.
    """
    root = S3Element.fromstring('BucketLocationConstraintResult', data)
    return root.text()

def _iso8601_to_localized_time(date_string):
    """
    Convert iso8601 date string into UTC time.

    :param date_string: iso8601 formatted date string.
    :return: :class:`datetime.datetime`
    """
    parsed_date = datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S.%fZ')
    localized_time = pytz.utc.localize(parsed_date)
    return localized_time

def parse_get_bucket_notification(data):
    """
    Parser for a get_bucket_notification response from S3.

    :param data: Body of response from get_bucket_notification.
    :return: Returns bucket notification configuration
    """
    root = S3Element.fromstring('GetBucketNotificationResult', data)

    notifications = _parse_add_notifying_service_config(
        root, {},
        'TopicConfigurations', 'TopicConfiguration'
    )
    notifications = _parse_add_notifying_service_config(
        root, notifications,
        'QueueConfigurations', 'QueueConfiguration'
    )
    notifications = _parse_add_notifying_service_config(
        root, notifications,
        'CloudFunctionConfigurations', 'CloudFunctionConfiguration'
    )

    return notifications

def _parse_add_notifying_service_config(data, notifications, service_key,
                                        service_xml_tag):

    arn_elt_name = NOTIFICATIONS_ARN_FIELDNAME_MAP[service_xml_tag]
    config = []
    for service in data.findall(service_xml_tag):
        config_item = {}
        config_item['Id'] = service.get_child_text('Id')
        config_item['Arn'] = service.get_child_text(arn_elt_name)
        config_item['Events'] = [
            event.text() for event in service.findall('Event')
        ]
        filter_terms = [
            {
                'Key': {
                    'FilterRules': [
                        {
                            'Name': xml_filter_rule.get_child_text('Name'),
                            'Value': xml_filter_rule.get_child_text('Value'),
                        }
                        for xml_filter_rule in xml_filter_rules.findall(
                                './S3Key/FilterRule')
                    ]
                }
            }
            for xml_filter_rules in service.findall('Filter')
        ]
        if len(filter_terms) > 0:
            config_item['Filter'] = filter_terms
        config.append(config_item)

    if len(config) > 0:
        notifications[service_key] = config

    return notifications

def parse_multi_object_delete_response(data):
    """Parser for Multi-Object Delete API response.

    :param data: XML response body content from service.

    :return: Returns list of error objects for each delete object that
    had an error.

    """
    root = S3Element.fromstring('MultiObjectDeleteResult', data)
    return [
        MultiDeleteError(errtag.get_child_text('Key'),
                         errtag.get_child_text('Code'),
                         errtag.get_child_text('Message'))
        for errtag in root.findall('Error')
    ]



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2017 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.fold_case_dict

This module implements a case insensitive dictionary.

:copyright: (c) 2017 by Minio, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

class FoldCaseDict(dict):
    def __init__(self, dictionary={}):
        self._data = self.__create(dictionary)

    def __create(self, value):
        if isinstance(value, dict):
            data = {}
            for k, v in value.items():
                if isinstance(v, dict):
                    data[k.lower()] = FoldCaseDict(self.__create(v))
                else:
                    data[k.lower()] = v
            return data
        else:
            return value

    def __getitem__(self, item):
        return self._data[item.lower()]

    def __contains__(self, item):
        return item.lower() in self._data

    def __setitem__(self, key, value):
        self._data[key.lower()] = self.__create(value)

    def __delitem__(self, key):
        del self._data[key.lower()]

    def __iter__(self):
        return (k for k in self._data.keys())

    def __len__(self):
        return len(self._data)

    def __eq__(self, other):
        if isinstance(other, dict):
            other = FoldCaseDict(other)
        elif isinstance(other, FoldCaseDict):
            pass
        else:
            raise NotImplementedError

        # Compare insensitively
        return self.items() == other.items()

    def __repr__(self):
        return str(self._data)

    def get(self, key, default=None):
        if not key.lower() in self:
            return default
        else:
            return self[key]

    def has_key(self, key):
        return key.lower() in self

    def items(self):
        return [(k, v) for k, v in self.iteritems()]

    def keys(self):
        return [k for k in self.iterkeys()]

    def values(self):
        return [v for v in self.itervalues()]

    def iteritems(self):
        for k, v in self._data.items():
            yield k, v

    def iterkeys(self):
        for k, v in self._data.items():
            yield k

    def itervalues(self):
        for k, v in self._data.items():
            yield v

    def update(self, dictionary):
        if not (isinstance(dictionary, dict) or
                isinstance(dictionary, FoldCaseDict)):
            raise TypeError

        for k, v in dictionary.items():
            self[k] = v

    def copy(self):
        copy_dict = FoldCaseDict()
        for k, v in self._data.items():
            copy_dict[k] = v
        return copy_dict

    def clear(self):
        self._data = {}

    def pop(self, key):
        return self._data.pop(key)

    def popitem(self):
        return self._data.popitem()



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015, 2016, 2017 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.error
~~~~~~~~~~~~~~~~~~~

This module provides custom exception classes for Minio library
and API specific errors.

:copyright: (c) 2015, 2016, 2017 by Minio, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""
from xml.etree import cElementTree
from xml.etree.cElementTree import ParseError

if hasattr(cElementTree, 'ParseError'):
    ## ParseError seems to not have .message like other
    ## exceptions. Add dynamically new attribute carrying
    ## value from message.
    if not hasattr(ParseError, 'message'):
        setattr(ParseError, 'message', ParseError.msg)
    _ETREE_EXCEPTIONS = (ParseError, AttributeError, ValueError, TypeError)
else:
    _ETREE_EXCEPTIONS = (SyntaxError, AttributeError, ValueError, TypeError)


class MinioError(Exception):
    """
    Base class for all exceptions

    :param message: User defined message.
    """
    def __init__(self, message, **kwargs):
        super(MinioError, self).__init__(**kwargs)
        self.message = message

    def __str__(self):
        return "{name}: message: {message}".format(
            name=self.__class__.__name__,
            message=self.message
        )

class InvalidEndpointError(MinioError):
    """
    InvalidEndpointError is raised when input endpoint URL is invalid.
    """
    pass


class InvalidBucketError(MinioError):
    """
    InvalidBucketError is raised when input bucket name is invalid.

    NOTE: Bucket names are validated based on Amazon S3 requirements.
    """
    pass


class InvalidArgumentError(MinioError):
    """
    InvalidArgumentError is raised when an unexpected
    argument is received by the callee.
    """
    pass


class InvalidSizeError(MinioError):
    """
    InvalidSizeError is raised when an unexpected size mismatch occurs.
    """
    pass


class InvalidXMLError(MinioError):
    """
    InvalidXMLError is raised when an unexpected XML tag or
    a missing tag is found during parsing.
    """
    pass


class MultiDeleteError(object):
    """
    Represents an error raised when trying to delete an object in a
    Multi-Object Delete API call :class:`MultiDeleteError <MultiDeleteError>`

    :object_name: Object name that had a delete error.
    :error_code: Error code.
    :error_message: Error message.
    """
    def __init__(self, object_name, err_code, err_message):
        self.object_name = object_name
        self.error_code = err_code
        self.error_message = err_message

    def __str__(self):
        string_format = '<MultiDeleteError: object_name: {} error_code: {}' \
                        ' error_message: {}>'
        return string_format.format(self.object_name,
                                    self.error_code,
                                    self.error_message)

class ResponseError(MinioError):
    """
    ResponseError is raised when an API call doesn't succeed.
    raises :exc:`ResponseError` accordingly.

    :param response: Response from http client :class:`urllib3.HTTPResponse`.
    """
    def __init__(self, response, method, bucket_name=None,
                 object_name=None):
        super(ResponseError, self).__init__(message='')
        # initialize parameter fields
        self._response = response
        self._xml = response.data
        self.method = method
        self.bucket_name = bucket_name
        self.object_name = object_name
        # initialize all ResponseError fields
        self.code = ''
        # Amz headers
        self.request_id = ''
        self.host_id = ''
        self.region = ''

        # handle the error
        self._handle_error_response(bucket_name)

    def get_exception(self):
        """
        Gets the error exception derived from the initialization of
        an ErrorResponse object

        :return: The derived exception or ResponseError exception
        """
        exception = known_errors.get(self.code)
        if exception:
            return exception(self)
        else:
            return self

    def _handle_error_response(self, bucket_name=None):
        """
        Sets error response uses xml body if available, otherwise
        relies on HTTP headers.
        """
        if not self._response.data:
            self._set_error_response_without_body(bucket_name)
        else:
            self._set_error_response_with_body(bucket_name)

    def _set_error_response_with_body(self, bucket_name=None):
        """
        Sets all the error response fields with a valid response body.
           Raises :exc:`ValueError` if invoked on a zero length body.

        :param bucket_name: Optional bucket name resource at which error
           occurred.
        :param object_name: Option object name resource at which error
           occurred.
        """
        if len(self._response.data) == 0:
            raise ValueError('response data has no body.')
        try:
            root = cElementTree.fromstring(self._response.data)
        except _ETREE_EXCEPTIONS as error:
            raise InvalidXMLError('"Error" XML is not parsable. '
                                  'Message: {0}'.format(error.message))
        for attribute in root:
            if attribute.tag == 'Code':
                self.code = attribute.text
            elif attribute.tag == 'BucketName':
                self.bucket_name = attribute.text
            elif attribute.tag == 'Key':
                self.object_name = attribute.text
            elif attribute.tag == 'Message':
                self.message = attribute.text
            elif attribute.tag == 'RequestId':
                self.request_id = attribute.text
            elif attribute.tag == 'HostId':
                self.host_id = attribute.text
        # Set amz headers.
        self._set_amz_headers()

    def _set_error_response_without_body(self, bucket_name=None):
        """
        Sets all the error response fields from response headers.
        """
        if self._response.status == 404:
            if bucket_name:
                if self.object_name:
                    self.code = 'NoSuchKey'
                    self.message = self._response.reason
                else:
                    self.code = 'NoSuchBucket'
                    self.message = self._response.reason
        elif self._response.status == 409:
            self.code = 'Conflict'
            self.message = 'The bucket you tried to delete is not empty.'
        elif self._response.status == 403:
            self.code = 'AccessDenied'
            self.message = self._response.reason
        elif self._response.status == 400:
            self.code = 'BadRequest'
            self.message = self._response.reason
        elif self._response.status == 301:
            self.code = 'PermanentRedirect'
            self.message = self._response.reason
        elif self._response.status == 307:
            self.code = 'Redirect'
            self.message = self._response.reason
        elif self._response.status in [405, 501]:
            self.code = 'MethodNotAllowed'
            self.message = self._response.reason
        elif self._response.status == 500:
            self.code = 'InternalError'
            self.message = 'Internal Server Error.'
        else:
            self.code = 'UnknownException'
            self.message = self._response.reason
        # Set amz headers.
        self._set_amz_headers()

    def _set_amz_headers(self):
        """
        Sets x-amz-* error response fields from response headers.
        """
        if self._response.headers:
            # keeping x-amz-id-2 as part of amz_host_id.
            if 'x-amz-id-2' in self._response.headers:
                self.host_id = self._response.headers['x-amz-id-2']
            if 'x-amz-request-id' in self._response.headers:
                self.request_id = self._response.headers['x-amz-request-id']
            # This is a new undocumented field, set only if available.
            if 'x-amz-bucket-region' in self._response.headers:
                self.region = self._response.headers['x-amz-bucket-region']

    def __str__(self):
        return ('ResponseError: code: {0}, message: {1},'
                ' bucket_name: {2}, object_name: {3}, request_id: {4},'
                ' host_id: {5}, region: {6}'.format(self.code,
                                                    self.message,
                                                    self.bucket_name,
                                                    self.object_name,
                                                    self.request_id,
                                                    self.host_id,
                                                    self.region))


# Common error responses listed here
# http://docs.aws.amazon.com/AmazonS3/latest/API/ErrorResponses.htmlRESTErrorResponses

class KnownResponseError(MinioError):
    def __init__(self, response_error, **kwargs):
        super(KnownResponseError, self).__init__(message=self.message, **kwargs)
        self.response_error = response_error

class AccessDenied(KnownResponseError):
    message = 'Access Denied'

class AccountProblem(KnownResponseError):
    message = 'There is a problem with your account that prevents the ' \
              'operation from completing successfully.'

class AmbiguousGrantByEmailAddress(KnownResponseError):
    message = 'The email address you provided is associated with ' \
              'more than one account.'

class BadDigest(KnownResponseError):
    message = 'The Content-MD5 you specified did not match what we received.'

class BucketAlreadyExists(KnownResponseError):
    message = 'The requested bucket name is not available. The ' \
              'bucket namespace is shared by all users of the system. ' \
              'Please select a different name and try again.'

class BucketAlreadyOwnedByYou(KnownResponseError):
    message = 'Your previous request to create the named bucket ' \
              'succeeded and you already own it.'

class BucketNotEmpty(KnownResponseError):
    message = 'The bucket you tried to delete is not empty.'

class CredentialNotSupported(KnownResponseError):
    message = 'This request does not support credentials.'

class CrossLocationLoggingProhibited(KnownResponseError):
    message = 'Cross-location logging not allowed. Buckets in one ' \
              'geographic location cannot log information to a bucket ' \
              'in another location.'

class EntityTooSmall(KnownResponseError):
    message = 'Your proposed upload is smaller than the minimum a' \
              'llowed object size.'

class EntityTooLarge(KnownResponseError):
    message = 'Your proposed upload exceeds the maximum allowed object size.'

class ExpiredToken(KnownResponseError):
    message = 'The provided token has expired.'

class IllegalVersioningConfigurationException(KnownResponseError):
    message = 'Indicates that the versioning configuration specified ' \
              'in the request is invalid.'

class IncompleteBody(KnownResponseError):
    message = 'You did not provide the number of bytes specified by the ' \
              'Content-Length HTTP header'

class IncorrectNumberOfFilesInPostRequest(KnownResponseError):
    message = 'POST requires exactly one file upload per request.'

class InlineDataTooLarge(KnownResponseError):
    message = 'Inline data exceeds the maximum allowed size.'

class InternalError(KnownResponseError):
    message = 'We encountered an internal error. Please try again.'

class InvalidAccessKeyId(KnownResponseError):
    message = 'The access key Id you provided does not exist in our records.'

class InvalidAddressingHeader(KnownResponseError):
    message = 'You must specify the Anonymous role.'

class InvalidArgument(KnownResponseError):
    message = 'Invalid Argument'

class InvalidBucketName(KnownResponseError):
    message = 'The specified bucket is not valid.'

class InvalidBucketState(KnownResponseError):
    message = 'The request is not valid with the current state of the bucket.'

class InvalidDigest(KnownResponseError):
    message = 'The Content-MD5 you specified is not valid.'

class InvalidEncryptionAlgorithmError(KnownResponseError):
    message = 'The encryption request you specified is not valid. ' \
              'The valid value is AES256.'

class InvalidLocationConstraint(KnownResponseError):
    message = 'The specified location constraint is not valid.'

class InvalidObjectState(KnownResponseError):
    message = 'The operation is not valid for the current state of the object.'

class InvalidPart(KnownResponseError):
    message = 'One or more of the specified parts could not be found. ' \
              'The part might not have been uploaded, or the specified ' \
              'entity tag might not have matched the part\'s entity tag'

class InvalidPartOrder(KnownResponseError):
    message = 'The list of parts was not in ascending order.Parts list ' \
              'must specified in order by part number.'

class InvalidPayer(KnownResponseError):
    message = 'All access to this object has been disabled.'

class InvalidPolicyDocument(KnownResponseError):
    message = 'The content of the form does not meet the conditions ' \
              'specified in the policy document.'

class InvalidRange(KnownResponseError):
    message = 'The requested range cannot be satisfied.'

class InvalidRequest(KnownResponseError):
    message = 'Invalid Request'

class InvalidSecurity(KnownResponseError):
    message = 'The provided security credentials are not valid.'

class InvalidSOAPRequest(KnownResponseError):
    message = 'The SOAP request body is invalid.'

class InvalidStorageClass(KnownResponseError):
    message = 'The storage class you specified is not valid.'

class InvalidTargetBucketForLogging(KnownResponseError):
    message = 'The target bucket for logging does not exist, ' \
              'is not owned by you, or does not have the appropriate ' \
              'grants for the log-delivery group.'

class InvalidToken(KnownResponseError):
    message = 'The provided token is malformed or otherwise invalid.'

class InvalidURI(KnownResponseError):
    message = 'Couldn\'t parse the specified URI.'

class KeyTooLong(KnownResponseError):
    message = 'Your key is too long.'

class MalformedACLError(KnownResponseError):
    message = 'The XML you provided was not well-formed ' \
              'or did not validate against our published schema.'

class MalformedPOSTRequest(KnownResponseError):
    message = 'The body of your POST request is not ' \
              'well-formed multipart/form-data.'

class MalformedXML(KnownResponseError):
    message = 'This happens when the user sends malformed xml (xml that ' \
              'doesn\'t conform to the published xsd) for the configuration.'

class MaxMessageLengthExceeded(KnownResponseError):
    message = 'Your request was too big.'

class MaxPostPreDataLengthExceededError(KnownResponseError):
    message = 'Your POST request fields preceding the ' \
              'upload file were too large.'

class MetadataTooLarge(KnownResponseError):
    message = 'Your metadata headers exceed the maximum allowed metadata size.'

class MethodNotAllowed(KnownResponseError):
    message = 'The specified method is not allowed against this resource'

class MissingAttachment(KnownResponseError):
    message = 'A SOAP attachment was expected, but none were found.'

class MissingContentLength(KnownResponseError):
    message = 'You must provide the Content-Length HTTP header.'

class MissingRequestBodyError(KnownResponseError):
    message = 'This happens when the user sends an empty xml document ' \
              'as a request. The error message is, "Request body is empty."'

class MissingSecurityElement(KnownResponseError):
    message = 'The SOAP 1.1 request is missing a security element.'

class MissingSecurityHeader(KnownResponseError):
    message = 'Your request is missing a required header.'

class NoLoggingStatusForKey(KnownResponseError):
    message = 'There is no such thing as a logging ' \
              'status subresource for a key.'

class NoSuchBucket(KnownResponseError):
    message = 'The specified bucket does not exist.'

class NoSuchKey(KnownResponseError):
    message = 'The specified key does not exist.'

class NoSuchLifecycleConfiguration(KnownResponseError):
    message = 'The lifecycle configuration does not exist.'

class NoSuchUpload(KnownResponseError):
    message = 'The specified multipart upload does not exist. ' \
              'The upload ID might be invalid, or the multipart \
              upload might have been aborted or completed.'

class NoSuchVersion(KnownResponseError):
    message = 'Indicates that the version ID specified in the ' \
              'request does not match an existing version.'

class APINotImplemented(KnownResponseError):
    message = 'A header you provided implies functionality ' \
              'that is not implemented.'

class NotSignedUp(KnownResponseError):
    message = 'Your account is not signed up.'

class NoSuchBucketPolicy(KnownResponseError):
    message = 'The specified bucket does not have a bucket policy.'

class OperationAborted(KnownResponseError):
    message = 'A conflicting conditional operation is currently in ' \
              'progress against this resource. Try again.'

class PermanentRedirect(KnownResponseError):
    message = 'The bucket you are attempting to access must be addressed ' \
              'using the specified endpoint. Send all future requests ' \
              'to this endpoint.'

class PreconditionFailed(KnownResponseError):
    message = 'At least one of the preconditions you specified did not hold.'

class Redirect(KnownResponseError):
    message = 'Temporary redirect.'

class RestoreAlreadyInProgress(KnownResponseError):
    message = 'Object restore is already in progress.'

class RequestIsNotMultiPartContent(KnownResponseError):
    message = 'Bucket POST must be of the enclosure-type multipart/form-data.'

class RequestTimeout(KnownResponseError):
    message = 'Your socket connection to the server was not read ' \
              'from or written to within the timeout period.'

class RequestTimeTooSkewed(KnownResponseError):
    message = 'The difference between the request time and the ' \
              'server\'s time is too large.'

class RequestTorrentOfBucketError(KnownResponseError):
    message = 'Requesting the torrent file of a bucket is not permitted.'

class SignatureDoesNotMatch(KnownResponseError):
    message = 'The request signature we calculated does not match the ' \
              'signature you provided.'

class ServiceUnavailable(KnownResponseError):
    message = 'Reduce your request rate.'

class SlowDown(KnownResponseError):
    message = 'Reduce your request rate.'

class TemporaryRedirect(KnownResponseError):
    message = 'You are being redirected to the bucket while DNS updates.'

class TokenRefreshRequired(KnownResponseError):
    message = 'The provided token must be refreshed.'

class TooManyBuckets(KnownResponseError):
    message = 'You have attempted to create more buckets than allowed.'

class UnexpectedContent(KnownResponseError):
    message = 'This request does not support content.'

class UnresolvableGrantByEmailAddress(KnownResponseError):
    message = 'The email address you provided does not match any account ' \
              'on record.'

class UserKeyMustBeSpecified(KnownResponseError):
    message = 'The bucket POST must contain the specified field name. ' \
              'If it is specified, check the order of the fields.'

known_errors = {
    'AccessDenied': AccessDenied,
    'AcccountProblem': AccountProblem,
    'AmbiguousGrantByEmailAddress': AmbiguousGrantByEmailAddress,
    'BadDigest': BadDigest,
    'BucketAlreadyExists': BucketAlreadyExists,
    'BucketAlreadyOwnedByYou': BucketAlreadyOwnedByYou,
    'BucketNotEmpty': BucketNotEmpty,
    'CredentialNotSupported': CredentialNotSupported,
    'CrossLocationLoggingProhibited': CrossLocationLoggingProhibited,
    'EntityTooSmall': EntityTooSmall,
    'EntityTooLarge': EntityTooLarge,
    'ExpiredToken': ExpiredToken,
    'IllegalVersioningConfigurationException': IllegalVersioningConfigurationException,
    'IncompleteBody': IncompleteBody,
    'IncorrectNumberOfFilesInPostRequest': IncorrectNumberOfFilesInPostRequest,
    'InlineDataTooLarge': InlineDataTooLarge,
    'InternalError': InternalError,
    'InvalidAccessKeyId': InvalidAccessKeyId,
    'InvalidAddressingHeader': InvalidAddressingHeader,
    'InvalidArgument': InvalidArgument,
    'InvalidBucketName': InvalidBucketName,
    'InvalidBucketState': InvalidBucketState,
    'InvalidDigest': InvalidDigest,
    'InvalidEncryptionAlgorithmError': InvalidEncryptionAlgorithmError,
    'InvalidLocationConstraint': InvalidLocationConstraint,
    'InvalidObjectState': InvalidObjectState,
    'InvalidPart': InvalidPart,
    'InvalidPartOrder': InvalidPartOrder,
    'InvalidPayer': InvalidPayer,
    'InvalidPolicyDocument': InvalidPolicyDocument,
    'InvalidRange': InvalidRange,
    'InvalidRequest': InvalidRequest,
    'InvalidSecurity': InvalidSecurity,
    'InvalidSOAPRequest': InvalidSOAPRequest,
    'InvalidStorageClass': InvalidStorageClass,
    'InvalidTargetBucketForLogging': InvalidTargetBucketForLogging,
    'InvalidToken': InvalidToken,
    'InvalidURI': InvalidURI,
    'KeyTooLong': KeyTooLong,
    'MalformedACLError': MalformedACLError,
    'MalformedPOSTRequest': MalformedPOSTRequest,
    'MalformedXML': MalformedXML,
    'MaxMessageLengthExceeded': MaxMessageLengthExceeded,
    'MaxPostPreDataLengthExceededError': MaxPostPreDataLengthExceededError,
    'MetadataTooLarge': MetadataTooLarge,
    'MethodNotAllowed': MethodNotAllowed,
    'MissingAttachment': MissingAttachment,
    'MissingContentLength': MissingContentLength,
    'MissingRequestBodyError': MissingRequestBodyError,
    'MissingSecurityElement': MissingSecurityElement,
    'MissingSecurityHeader': MissingSecurityHeader,
    'NoLoggingStatusForKey': NoLoggingStatusForKey,
    'NoSuchBucket': NoSuchBucket,
    'NoSuchKey': NoSuchKey,
    'NoSuchLifecycleConfiguration': NoSuchLifecycleConfiguration,
    'NoSuchUpload': NoSuchUpload,
    'NoSuchVersion': NoSuchVersion,
    'NotImplemented': APINotImplemented,
    'NotSignedUp': NotSignedUp,
    'NoSuchBucketPolicy': NoSuchBucketPolicy,
    'OperationAborted': OperationAborted,
    'PermanentRedirect': PermanentRedirect,
    'PreconditionFailed': PreconditionFailed,
    'Redirect': Redirect,
    'RestoreAlreadyInProgress': RestoreAlreadyInProgress,
    'RequestIsNotMultiPartContent': RequestIsNotMultiPartContent,
    'RequestTimeout': RequestTimeout,
    'RequestTimeTooSkewed': RequestTimeTooSkewed,
    'RequestTorrentOfBucketError': RequestTorrentOfBucketError,
    'SignatureDoesNotMatch': SignatureDoesNotMatch,
    'ServiceUnavailable': ServiceUnavailable,
    'SlowDown': SlowDown,
    'TemporaryRedirect': TemporaryRedirect,
    'TokenRefreshRequired': TokenRefreshRequired,
    'TooManyBuckets': TooManyBuckets,
    'UnexpectedContent': UnexpectedContent,
    'UnresolvableGrantByEmailAddress': UnresolvableGrantByEmailAddress,
    'UserKeyMustBeSpecified': UserKeyMustBeSpecified,
}



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 Compatible Cloud Storage, (C) 2018 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.sse
~~~~~~~~~~~~~~~~~~~

This module contains core API parsers.

:copyright: (c) 2018 by Minio, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""
import base64
import hashlib
import json
from .error import (InvalidArgumentError, InvalidSizeError, InvalidXMLError, NoSuchBucketPolicy)

class SSE_C(object):

    def __init__(self, key):
        self.key = key
        if len(self.key) != 32:
            raise InvalidSizeError("SSE-C keys need to be 256 bit base64 encoded")
    
    def type(self):
        return "SSE-C"
    
    def marshal(self):

        b64key = base64.b64encode(self.key)
        md5 = hashlib.md5()
        md5.update(self.key)
        md5_key = base64.b64encode(md5.digest()).decode()
        keys = {
            "X-Amz-Server-Side-Encryption-Customer-Algorithm": "AES256",
            "X-Amz-Server-Side-Encryption-Customer-Key": b64key.decode(),
            "X-Amz-Server-Side-Encryption-Customer-Key-MD5": md5_key
            }

        return keys

class copy_SSE_C(object):

    def __init__(self, key):
        self.key = key
        if len(self.key) != 32:
            raise InvalidArgumentError("Length of Customer key must be 32 Bytes")

    def type(self):
        return "copy_SSE-C"

    def marshal(self):
        b64key = base64.b64encode(self.key)
        md5 = hashlib.md5()
        md5.update(self.key)
        md5_key = base64.b64encode(md5.digest()).decode()
        keys = {
            "X-Amz-Copy-Source-Server-Side-Encryption-Customer-Algorithm":"AES256",
            "X-Amz-Copy-Source-Server-Side-Encryption-Customer-Key": b64key.decode(),
            "X-Amz-Copy-Source-Server-Side-Encryption-Customer-Key-MD5": md5_key
        }
        return keys


class SSE_KMS(object):
    def __init__(self, key, context):
        self.key = key
        self.context = context

    def type(self):
        return "SSE-KMS"

    def marshal(self):
        keys = {
            "X-Amz-Server-Side-Encryption-Aws-Kms-Key-Id": self.key,
            "X-Amz-Server-Side-Encryption":"aws:kms"
        }

        if self.context:
                ctx_str = json.dumps(self.context)
                ctx_str = bytes(ctx_str, 'utf-8')
                b64key = base64.b64encode(ctx_str)
                header = {"X-Amz-Server-Side-Encryption-Context": b64key.decode()}
                keys.update(header)
        
        return keys
        

class SSE_S3(object):
    def type(self):
        return "SSE-S3"

    def marshal(self):
        keys = {
            "X-Amz-Server-Side-Encryption":"AES256"
        }
        return keys
    



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2017 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.thread_pool
~~~~~~~~~~~~

This module implements a thread pool API to run several tasks
in parallel. Tasks results can also be retrieved.

:copyright: (c) 2017 by Minio, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from threading import Thread
from .compat import queue


class Worker(Thread):
    """ Thread executing tasks from a given tasks queue """

    def __init__(self, tasks_queue, results_queue, exceptions_queue):
        Thread.__init__(self)
        self.tasks_queue = tasks_queue
        self.results_queue = results_queue
        self.exceptions_queue = exceptions_queue
        self.daemon = True
        self.start()

    def run(self):
        """ Continously receive tasks and execute them """
        while True:
            task = self.tasks_queue.get()
            if task is None:
                self.tasks_queue.task_done()
                break
            # No exception detected in any thread,
            # continue the execution.
            if self.exceptions_queue.empty():
                try:
                    # Execute the task
                    func, args, kargs = task
                    result = func(*args, **kargs)
                    self.results_queue.put(result)
                except Exception as e:
                    self.exceptions_queue.put(e)
            # Mark this task as done, whether an exception happened or not
            self.tasks_queue.task_done()

class ThreadPool:
    """ Pool of threads consuming tasks from a queue """

    def __init__(self, num_threads):
        self.results_queue = queue()
        self.exceptions_queue = queue()
        self.tasks_queue = queue(num_threads)
        self.num_threads = num_threads

    def add_task(self, func, *args, **kargs):
        """ Add a task to the queue """
        self.tasks_queue.put((func, args, kargs))

    def start_parallel(self):
        """ Prepare threads to run tasks"""
        for _ in range(self.num_threads):
            Worker(self.tasks_queue, self.results_queue, self.exceptions_queue)

    def result(self):
        """ Stop threads and return the result of all called tasks """
        # Send None to all threads to cleanly stop them
        for _ in range(self.num_threads):
            self.tasks_queue.put(None)
        # Wait for completion of all the tasks in the queue
        self.tasks_queue.join()
        # Check if one of the thread raised an exception, if yes
        # raise it here in the function
        if not self.exceptions_queue.empty():
            raise self.exceptions_queue.get()
        return self.results_queue

